(in-package :araneida)

;; that portion of the request path after base-url
(defmethod request-path-info ((r request))
  (let ((path (url-path (request-url r)))
        (ppath (url-path (request-base-url r))))
    (subseq path (length ppath) nil)))

(defgeneric request-header ((r request) name)
  (:documentation "Returns a list containing the values of all header lines in the request R given by the keyword NAME"))

(defmethod request-header ((r request) name)
  (cdr (assoc name (request-headers r) :test #'string=)))

(defmethod request-if-modified-since ((request request)
                                      &optional (default nil))
  "Retrieve and parse the date in the If-Modified-Since header field.  Return D
EFAULT if the header is absent or unparseable"
  (let ((if-mod-since (car (request-header request :if-modified-since))))
    (if if-mod-since
             (or (date:parse-time (car (split if-mod-since 2 '(#\;))))
                 default)
      default)))

(defmethod request-cookie ((request request) name)
  (labels ((same-name-p (st)
	     (loop
	      (unless (and (> (length st) 1)
			   (char= (elt st 0) #\Space)) (return))
	      (setf st (subseq st 1)))
	     (let ((l (length name)))
               (and (> (length st) l)
                    (string= st name :end1 l)
                    (eql (char st l) #\=)))))
  (let ((cookies (car (request-header request :cookie))))
    (loop for vv in (split cookies nil '(#\;))
          if (same-name-p vv)
          return (second (split vv 2 '(#\=)))))))


;; this takes an alist not a request, hence the name
(defun body-param (name alist)
  "Look in the request body ALIST for the value of the parameter NAME"
  (cadr (assoc name alist :test #'string=)))

(defun body-params (name alist &key (case-sensitive nil))
  "Look in the request body ALIST for the values of the parameters starting NAME, returning a list of KEY VALUE pairs"
  ;; find all parameters starting NAME
(remove-if-not
   (if case-sensitive
       (lambda (x)
         (destructuring-bind (k v) x
           (and (>= (length k) (length name))
                (string= k name :end1 (length name)))))
     (lambda (x)
       (destructuring-bind (k v) x
         (and (>= (length k) (length name))
              (string-equal k name :end1 (length name))))))
   alist))


(defmethod dispatch-request ((request request) handlers &optional discriminator)
  "Find the best match for REQUEST in the list HANDLERS"
  (unless discriminator (setf discriminator (request-url request)))
  (destructuring-bind
        (method match prefix func &optional needs-discriminator)
      (find-export (urlstring discriminator) handlers (request-method request))
    (unless method (return-from dispatch-request nil))
    (setf (request-base-url request) (parse-urlstring prefix))
    (let ((rest-of-url
           (subseq (urlstring discriminator)
                   (length (urlstring (request-base-url request)))
                   nil)))
      (cond ((and needs-discriminator (consp func))
             (apply (car func) request handlers discriminator
                    rest-of-url (cdr func)))
            ((consp func)
             (apply (car func) request rest-of-url (cdr func)))
            (needs-discriminator
             (funcall func request handlers discriminator rest-of-url))
            (t
             (funcall func request rest-of-url))))))

;;; rfc 1945 p26
(defvar *http-error-codes*
  '((400 . "Bad Request")
    (401 . "Unauthorized")
    (403 . "Forbidden")
    (404 . "Not Found")
    (500 . "Internal Server Error")
    (501 . "Not Implemented")
    (502 . "Bad Gateway")
    (503 . "Service Unavailable")))

(defmethod request-send-headers ((request request) &key
                                 (content-type "text/html")
                                 content-length
                                 expires
				 cache-control
				 location
				 pragma
				 set-cookie
                                 (last-modified (get-universal-time))
                                 (response-text "OK")
                                 (response-code 200))
  "Send HTTP/1.0 headers in response to REQUEST"
  (let ((stream (request-stream request))
        (eol #.(format nil "~%"))
        (date (date:universal-time-to-rfc-date (get-universal-time)))
        (lmtime (date:universal-time-to-rfc-date last-modified)))
    (format stream "HTTP/1.0 ~D ~A
Date: ~A
Server: ~A
Connection: close
Last-Modified: ~A
Content-type: ~A~%"
            response-code response-text date
            *araneida-product-tokens* lmtime content-type)
    (if content-length
        (princ (s. "Content-length: " content-length eol) stream))
    (if set-cookie
	(let ((cookies (if (listp set-cookie) set-cookie (list set-cookie))))
	  (dolist (cookie cookies)
	    (princ (s. "Set-Cookie: " cookie eol) stream))))
    (if cache-control
        (princ (s. "Cache-Control: " cache-control eol) stream))
    (if location
        (princ (s. "Location: " location eol) stream))
    (if pragma
        (princ (s. "Pragma: " pragma eol) stream))
    (if expires
        (princ (s. "Expires: "
		   (if (numberp expires)
		       (date:universal-time-to-rfc-date expires)
		       expires)
		   eol)
               stream))
    (terpri stream)))
    
  
(defmethod request-send-error ((request request) error-code &rest extra-stuff)
  "Send the client HTTP headers and HTML body for an error message
with numeric code ERROR-CODE.  EXTRA-STUFF is an optional format
control and arguments for additional (PRE-formatted) information to
put on the page"
  (let ((stream (request-stream request))
        (extra (if extra-stuff (apply #'format nil extra-stuff)))
        (error-text (cdr (assoc error-code *http-error-codes*))))
    (request-send-headers request
                          :response-code error-code :response-text error-text)
    (format stream
            (html `(html (head (title ,(s. error-code) " " ,error-text))
                         (body
                          (h2 ,error-text)
                          (p "Was asked for URL "
                             (tt ,(urlstring (request-url request)))
                             ", but it didn't happen for us.  Sorry")
                          ,@(when extra
                              `((H3 "Additional information: ")
                                (pre ,(html-escape extra))))))))))

(defmethod request-redirect ((request request) new-url &rest headers)  
  (let ((urlstring (urlstring
                    (if (typep new-url 'url)
			new-url
			(merge-url (request-url request) new-url)))))
    (apply #'request-send-headers
	   request
	   :location urlstring
	   :expires "Fri, 30 Oct 1998 14:19:41 GMT"
	   :pragma "no-cache"
	   :response-code 302 :response-text "Redirected"
	   headers)
    (format (request-stream request)
	    "~%<h1>Redirected</h1><p>Continue <a href=\"~A\">~A</a>"
	    urlstring urlstring)))


